################################################
#
# Pillars of Eternity 1 & 2
#    Version 0.2
#
#  Installation:
#   * Download base Console - https://github.com/TheHologram/unity-console/releases
#   * Unpack zip to game folder
#   * Drag game exe over IPA.exe to instrument plugins. (Modifies game dlls)
#   * Copy this file to <game>\Plugins\Console\Lib\
#   * Start Game
#   * Use Ctrl+` (Ctrl+BackTick) to open console
#   * Type: 
#     >>> import pillars
#     >>> pillars.showWindow()
#   
################################################

def unity(func):
    """ Decorator for wrapping unity calls """
    import functools
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            import coroutine
            if args == None: args=()
            return coroutine.start_new_coroutine(func, args, kwargs)
        except Exception:
            pass
    return wrapper   
    

def cleanup():
    import unity_util
    unity_util.clean_behaviors()
    
def clear():
    keys = [key for key in globals().keys() if not key.startswith('__')]
    for key in keys:
        del globals()[key]
        
def cls():
    from System import Console
    Console.Clear()
    
def appName():
    global _appname
    try:
        _appname
    except NameError:
        _appname = ''
        import ctypes
        buffer = ctypes.create_string_buffer(260)
        ctypes.windll.kernel32.GetModuleFileNameA(None, buffer, len(buffer))
        _appname = buffer.value
    return _appname
    
def appFileName():
    from System.IO import Path
    return Path.GetFileNameWithoutExtension(appName())
    
def isPillars2():
    global _ispillars2
    try:
        _ispillars2
    except NameError:
        _ispillars2 = appFileName().lower() in ('pillarsofeternity2', 'pillarsofeternityii')
    return _ispillars2

class HorizontalScope():
    def __init__(self, options=None):
        self.options = options
        
    def __enter__(self):
        from UnityEngine import GUILayout, GUILayoutOption
        import System
        if self.options:
            GUILayout.BeginHorizontal(System.Array[GUILayoutOption](self.options))
        else:
            GUILayout.BeginHorizontal()
        return self
    def __exit__(self, exc_type, exc_value, traceback):
        from UnityEngine import GUILayout, GUILayoutOption
        GUILayout.EndHorizontal()
        
class VerticalScope():
    def __init__(self,text=None,style=None,options=None):
        self.text=text
        self.style=style
        self.options = options
    def __enter__(self):
        from UnityEngine import GUILayout, GUILayoutOption
        import System
        if self.options:
            GUILayout.BeginVertical(System.Array[GUILayoutOption](self.options))
        elif self.text and self.style:
            GUILayout.BeginVertical(self.text,self.style)
        else:
            GUILayout.BeginVertical(self.text,self.style)
        return self
    def __exit__(self, exc_type, exc_value, traceback):
        from UnityEngine import GUILayout, GUILayoutOption
        GUILayout.EndVertical()
    
def getFrontWindow():
    try:
        if isPillars2():
            from Game.UI import UIWindowManager
        else:
            import UIWindowManager
            
        mgr = UIWindowManager.Instance
        windows = mgr._UIWindowManager__m_Windows
        for w in windows:
            if w and w.WindowActive():
                return w
    except:
        pass
    return None
    
def getActiveWindows():
    try:
        if isPillars2():
            from Game.UI import UIWindowManager
        else:
            import UIWindowManager
        mgr = UIWindowManager.Instance
        windows = mgr._UIWindowManager__m_Windows
        return [w for w in windows if w and w.WindowActive()]
    except:
        pass
    return None
    
def getWindowsOfType(type):
    try:
        if isPillars2():
            from Game.UI import UIWindowManager
        else:
            import UIWindowManager
            
        mgr = UIWindowManager.Instance
        windows = mgr._UIWindowManager__m_Windows
        return [w for w in windows if w and w.WindowActive() and w.GetType() == type]
    except:
        pass
    return None

    
def trimText(text):
    global _nohtml
    try:
        _nohtml
    except NameError:
        from System.Text.RegularExpressions import Regex
        _nohtml = Regex(r"(?:<([^>]+)>)|(?:\[([^\]]+)\])")
    return _nohtml.Replace(text,'')

def getDisplayName(item):
    if isPillars2():
        from Game import CharacterStatsUtility
        return trimText(CharacterStatsUtility.Name(item))
    else:
        return item.DisplayName.GetText()

def getName(item):
    if isPillars2():
        from Game import CharacterStatsUtility
        return trimText(CharacterStatsUtility.Name(item))
    else:
        return item.Name.GetText()
     
def getPartyMembers(minions=False):
    if isPillars2():
        from Game import PartyManager
        for pm in PartyManager.Instance.GetActivePartyMembers():
            rai = pm.AIController
            if rai == None: continue
            cs = rai.CharacterStats
            yield (pm, cs, rai)
                
    else:
        import UIPartyPortraitBar
        portraits = UIPartyPortraitBar.Instance._UIPartyPortraitBar__m_Portraits
        showAll = False
        for p in portraits:
            cs = p.SelectedCharacter 
            if cs == None: continue
            ai = p.PartyMemberAI
            yield (p, cs, ai)
            
            if minions:
                m = p.Minion
                if m != None:
                    cs = m.SelectedCharacter
                    ai = m.SelectedPartyMemberAI
                    if cs != None: 
                        yield (m, cs, ai)

def Heal(cs, ai):
    if not cs or not ai: return
    try:
        if isPillars2():
            from Game import Health
            from Game.GameData import GlobalGameSettingsGameData, GlobalReferencesGameData, CharacterClass
        else:
            import Health
            
        health = ai.GetComponent[Health]()
        
        if isPillars2():
            trait = cs.GetAccruedResourceTrait(CharacterClass.Cipher)
            if trait: trait.Focus = trait.GetResourceMax().Value
            
            trait = cs.GetAccruedResourceTrait(CharacterClass.Chanter)
            if trait: trait.PhraseCount = trait.GetResourceMax()
            
            #trait = cs.GetAccruedResourceTrait(CharacterClass.Monk)
            #if trait: trait.ClearAllWounds()
            trait = cs.GetAccruedResourceTrait(CharacterClass.Monk)
            if trait:
                while not trait.IsResourceAtMax():
                    trait.AddNewWound()
            
            if health.IsDeadOrUnconscious: 
                health.Revive(cs.gameObject)
            if health.CountInjuries() > 0:
                for keyword in GlobalReferencesGameData.Instance.InjuryKeywords:
                    cs.StatusEffectManager.ClearStatusEffectsByKeyword(keyword, True)

            health.CurrentHealth = health.MaxHealth
            
            globalSettings = GlobalGameSettingsGameData.Instance
            cs.ResetClassAbilityPoolPoints()   
            cs.ResetSpellCastCounts()
            cs.RecoveryTimer = 0
            cs.CurrentGrimoireCooldown = 0
            

        else:
            trait = cs.FindFocusTrait()
            if trait != None: trait.Focus = trait.MaxFocus
            if health.Unconscious: health.WakeUp()
            elif health.Dead: health.OnRevive()

            health.HealthPercentage = 1.0
            health.StaminaPercentage = 1.0
            cs.CurrentFatigueLevel = 0
            
            if cs.ActiveStatusEffects.Count > 0:
                cs.ClearEffectsFromAfflictions()
            if cs.RecoveryTimer > 0:
                cs.RecoveryTimer = 0.0
                for a in cs.ActiveAbilities:
                    a.RestoreCooldown()
                    
            if cs.SpellCastCount.Length > 0:
                for i in xrange(0,cs.SpellCastCount.Length):
                    cs.SpellCastCount[i] = 0
    except Exception,e:
        print(str(e))

def AutoHeal():                
    for p, cs, ai in getPartyMembers(minions=True):
        Heal(cs, ai)

def TeleportPartyToDestination():
    if isPillars2():
        from Game import CameraControl, Mover
        from Game.AI import MoveParams
        
        first = True
        for pm, cs, rai in getPartyMembers(minions=True):
            if rai == None: continue
            moved = False
            behavior = rai.BehaviorStack.CurrentBehavior
            for i in xrange(rai.BehaviorStack.BehaviorCount):
                b = rai.BehaviorStack.GetBehavior(i)
                if b == None: continue
                params = b.AIParams
                if params == None: continue
                if params.GetType() == MoveParams:
                    rai.transform.position = params.GetPosition()
                    moved = True
            if not moved and rai.IsWaitingToMove():
                pm.transform.position = rai.Mover.Goal                    
                moved = True
                
            if moved and first:
                CameraControl.Instance.FocusOnPoint(rai.transform.position)
                first = False
        
    else:
        import PartyMemberAI, AI, CameraControl
        
        first = True
        for rai in PartyMemberAI.PartyMembers:
            if rai != None and rai.DestinationCircleState.GetType() == AI.Player.Move:
                rai.transform.position = rai.DestinationCirclePosition
                if first:
                    CameraControl.Instance.FocusOnPoint(rai.transform.position)
                    first = False
            
def addModToItem(item):
    import unity_util
    import UnityEngine
    import math
    from UnityEngine import GUI, GUILayout, GUIStyle, GUIUtility, Screen, Rect, Vector3, Input, KeyCode, Vector2
    from UnityEngine import Event, EventType, WaitForSeconds, GameObject, GUILayoutOption, GUIContent
    import System
    
    if isPillars2():
        from Game import Equippable, ItemMod, StatusEffectFormatMode
    else:
        import Equippable, ItemMod
               
    class AddModWindow():
        def __init__(self):
            self.gameObject = None # will be assigned if exists as member
            self.component = None # will be assigned if exists as member
            self.close = False
            width = 250 if isPillars2() else 300
            height = 650
            cwidth, cheight= (Screen.width-width)/2, (Screen.height-height)/2
            self.windowRect = Rect (0, cheight, width, height)
            self.drawWindowRect = Rect (0, 0, width, 50)
            self.scrollpos = Vector2(0,0)
            self.allLoaded = False
                        
            # load default (what is in memory)
            self.mods = {}
            items = []
            
            from System.Text.RegularExpressions import Regex
            nohtml = Regex(r"(?:<([^>]+)>)|(?:\[([^\]]+)\])")
            
            if isPillars2():
                
                def getModData(x):
                    return x.ItemModData
                def getModName(x):
                    return getName(getModData(x))
                def getTooltip(x):
                    return trimText(x.GetEffects(StatusEffectFormatMode.Tooltip))
            else:
                def getModData(x):
                    return x.Mod
                def getModName(x):
                    return x.Mod.DisplayName.GetText()
                def getTooltip(x):
                    return trimText(x.Mod.GetTooltipContent(None))
            
            if isPillars2():
                from Onyx import OnyxObject
                def isOfType(x, y):
                    itemtype = x.GetType()
                    return itemtype == Equippable or itemtype.IsSubclassOf(Equippable)
                items = [x.Value.Target for x in OnyxObject._OnyxObject__s_allObjects if x.Value.IsAlive and isOfType(x.Value.Target, Equippable)]
               
            else:
                items = UnityEngine.Object.FindObjectsOfType[Equippable]()
                
            for equipItem in items: 
                for mod in equipItem.AttachedItemMods:
                    name = getModName(mod)
                    if not name in self.mods: 
                        self.mods[name] = mod
                        #print(name, getTooltip(mod))
            self.keys = sorted(self.mods)

            if isPillars2():
                self.allLoaded = True
            else:
                def loadall():
                    import GameResources
                    names = ['acceptance', 'accurate1', 'accurate2', 'accurate3', 'accurate4', 'accurate6', 'accurate_armor_1', 'accurate_armor_2', 'adraswen', 'airtight', 'ambushing', 'amenity', 'anger', 'annihilation', 'athletic', 'banishing', 'bargaining', 'bonusmeleedamagefromwounds', 'bonusrangeddamage', 'breakout', 'burndtbonus', 'burninglash', 'burnproofed', 'constantrecoverybonus', 'coordinating', 'corrodedtbonus', 'corrodeproofed', 'corrosivelash', 'crushinglash', 'crushproofed', 'damaging1', 'damaging2', 'damaging3', 'dazzling', 'deadmanstands', 'decreasecrits', 'defiant', 'deflector', 'deflector2', 'denial', 'depression', 'disengagementdefensebonus', 'disorienting', 'draining', 'dyingboar', 'exceptional_armor', 'exceptional_weapon', 'false_life', 'fenwalkers', 'fine_armor', 'fine_weapon', 'finishingblowbonus', 'fireballnecklace', 'fortitudebonus', 'fortitudebonus2', 'freedom', 'freezedamageregen', 'freezedtbonus', 'freezeproofed', 'freezinglash', 'grefs_authority', 'guarding', 'healingbonus', 'immuneboost', 'immunity_confused_mod', 'immunity_stuck_mod', 'inaccurate', 'increasechanteraoe', 'increasedconstantrecovery', 'interfering', 'invigorating', 'irritating', 'liberating', 'loyal', 'majorstaminabonus', 'marking', 'martyr', 'mechanic', 'meleedamagebonus', 'meleedamagebonus5', 'mending', 'movementbonus', 'noisy', 'obstructing', 'ofconcentration10', 'ofconcentration25', 'ofconcentration5neg', 'ofconstitution1', 'ofconstitution2', 'ofconstitution3', 'ofcriticalhitmultbonus', 'ofdeflection10', 'ofdeflection15', 'ofdeflection5', 'ofdexterity1', 'ofdexterity2', 'ofdexterity3', 'offortitude10', 'offortitude15', 'offortitude18', 'offortitude5', 'ofintellect1', 'ofintellect1neg', 'ofintellect2', 'ofintellect2above50stamina', 'ofintellect3', 'ofmight1', 'ofmight2', 'ofmight3', 'ofperception1', 'ofperception2', 'ofperception3', 'ofreflexes10', 'ofreflexes15', 'ofreflexes5', 'ofresolve1', 'ofresolve2', 'ofresolve3', 'ofwill10', 'ofwill15', 'ofwill3neg', 'ofwill5', 'overbearing', 'overseeing', 'piercedtbonus', 'pierceproofed', 'piercinglash', 'predatory', 'pronereduction', 'px1_time_siphon', 'px1_weapon_durgan-refined', 'px4_spellstriking_firebrand', 'ragebonus', 'rangeddeflectionbonus', 'rawlash', 'reflexbonus', 'reflexbonus2', 'regeneration', 'reliable', 'rending', 'ring_of_eternal_funding_mod', 'sealoffaith', 'secondchance', 'shockdtbonus', 'shockinglash', 'shockproofed', 'shrouded', 'sickening', 'slashdtbonus', 'slashinglash', 'slashproofed', 'slaying_beast', 'slaying_hominid', 'slaying_primordial', 'slaying_spirit', 'slaying_vessel', 'slaying_wilder', 'sneakbonus', 'soothing', 'soulwhipbonus', 'speed', 'speed10', 'speed30', 'spellbind_minor_firebrand', 'spellchance_dominate', 'spellchance_restore_spiritshift', 'spelldamagemult', 'spelldefense_holy_meditation', 'spelldefensebonus', 'spiderfingers', 'stalkerslinkbonus', 'staminabonus', 'stealth_aura', 'stealthy', 'stunning', 'sturdy', 'superb_armor', 'superb_weapon', 'survivor', 'terrible_weapon', 'torchflame', 'unconquerable', 'valiant', 'vicious', 'wary', 'wildstrikebonus', 'wizardrya', 'wizardryb', 'wounding', 'wreathing_flame_aura', 'zealouscommandbonus']
                    for name in names:
                        mod = GameResources.LoadPrefab[ItemMod](name, False)
                        if mod: 
                            name = getModName(mod)
                            if not name in self.mods: 
                                self.mods[name] = mod
                    self.keys = sorted(self.mods)
                    self.allLoaded = True
            
            def FuncWindowGUI(windowid):            
                try:
                    with HorizontalScope():
                        GUILayout.Label(getDisplayName(item))
                        GUILayout.FlexibleSpace()
                        if GUILayout.Button('X'):
                            self.close = True
                            return
                    
                    options = System.Array[GUILayoutOption]([GUILayout.Width(width-20), GUILayout.Height(500)])
                    self.scrollpos = GUILayout.BeginScrollView(self.scrollpos, GUI.skin.window, options)
                    for key in self.keys:
                        with HorizontalScope():
                            mod = self.mods[key]
                            tooltip = getTooltip(mod)
                            GUILayout.Label(GUIContent(key, tooltip))
                            GUILayout.FlexibleSpace()
                            if GUILayout.Button(GUIContent('+', tooltip)):
                                item.AttachItemMod(getModData(mod))
                    GUILayout.EndScrollView()
                    
                    if not self.allLoaded:
                        with HorizontalScope():
                            GUILayout.Label('Load All Mods (Caution)')
                            GUILayout.FlexibleSpace()
                            if GUILayout.Button('Load'):
                                loadall()                           
                    
                    if GUI.tooltip:
                        GUILayout.Label(GUI.tooltip)
                    else:
                        GUILayout.Label("Note: Only mods already on other items are available")
                    
                except Exception, e:
                    GUILayout.Label("Error: " + str(e))
                    
                GUI.DragWindow(self.drawWindowRect)
            self.windowCallback = GUI.WindowFunction(FuncWindowGUI)
                      
        def OnGUI(self):
            if self.close: 
                pass
                # obj = self.gameObject
                # if obj:
                    # self.gameObject = None
                    # UnityEngine.Object.DestroyObject(obj)
                    # return
            else:
                self.windowRect = GUI.Window(2, self.windowRect, self.windowCallback, 'Item Window', GUI.skin.window)
    return AddModWindow()
    #return unity_util.create_gui_behavior(Controller)
    

def showWindow(visible=True):
    import unity_util
    import UnityEngine
    import math
    from UnityEngine import GUI, GUILayout, GUIStyle, GUIUtility, Screen, Rect, Vector3, Input, KeyCode, Vector2
    from UnityEngine import Event, EventType, WaitForSeconds, GameObject, GUIContent, GUILayoutOption, GUISkin, Font
    import System
    from System.Reflection import BindingFlags
    from System.Text.RegularExpressions import Regex
    
    if isPillars2():
        from Game import GameState, PlayerInventory, Health, Equippable, ReputationManager, Disposition, CharacterStatsUtility, StatusEffectFormatMode, ShipCrewManager, SevereInjuryManager, PartyManager, SpecialCharacterInstanceID, CharacterStatsMutable, InstanceID, CompanionInstanceID
        from Game.UI import UIWindowManager, UIHudWindow, UIPartyPortraitBar, UICraftingManager, UIItemInspectManager, UIInventoryManager, UIShipManagementManager
        import Game.Scripts 
        from Game.GameData import WindowType, CharacterClass
        from Game.GameData import RecipeData as Recipe
        from Game.GameData import Axis
        from Onyx import OnyxInt
        import ComponentUtils
    else:
        import GameState, PlayerInventory, Health, Equippable, ReputationManager, Disposition, Recipe
        import UIWindowManager, UIHudWindow, UIPartyPortraitBar, UICraftingManager, UIItemInspectManager, UIInventoryManager
        import StrongholdEvent
        from UIHudWindow import WindowType
        
    nohtml = Regex(r"(?:<([^>]+)>)|(?:\[([^\]]+)\])")
    windowWidth = 250 if isPillars2() else 300
                       
    unity_util.clean_behaviors()
    
    def CloneSkin(skin):
        newskin = GUISkin()
        props = newskin.GetType().GetProperties(BindingFlags.Public | BindingFlags.Instance| BindingFlags.SetProperty)
        for prop in [x.Name for x in props]:
            try:
                setattr(newskin, prop, getattr(skin, prop))
            except:
                pass
        return newskin
    
    
    class Controller():
        def __init__(self):
            self.gameObject = None # will be assigned if exists as member
            self.component = None # will be assigned if exists as member
            self.drawWindowRect = Rect (0, 0, 150, 50)
            self.skin = None

            def getGlobalState():
                global windowState
                try:
                    return windowState
                except NameError:
                    class Expando(object):
                        pass

                    windowState = Expando()
                    windowState.visible = True
                    windowState.windowRect = Rect (Screen.width - windowWidth, 0, windowWidth, Screen.height)
                    windowState.minimize = False
                    windowState.showStats = False
                    windowState.showSkills = False
                    windowState.showActions = False
                    windowState.showMultipliers = False
                    windowState.showGlobal = False
                    windowState.showGlobalActions = False
                    windowState.showDisposition = False
                    windowState.showFactions = False
                    windowState.showCompanionDisp = False
                    windowState.factionScroll = Vector2(0,0)
                    windowState.compDispScroll = Vector2(0,0)
                    windowState.autoCmds = False
                    windowState.runAutoCmds = False
                    windowState.autoHeal = True
                    
                    windowState.pinvscrollpos = Vector2(0,0)
                    windowState.sinvscrollpos = Vector2(0,0)
                    windowState.finvscrollpos = Vector2(0,0)
                    windowState.dinvscrollpos = Vector2(0,0)
                    windowState.cinvscrollpos = Vector2(0,0)
                    windowState.showPlayerInventory = True
                    windowState.showStashInventory = False
                    windowState.showFoodInventory = False
                    windowState.showDrinkInventory = False
                    windowState.showCrewInventory = False
                    
                    return windowState
            
            self.state = getGlobalState()
            self.state.visible = visible
            self.timerRunning = False
            self.counter = 0
            self.addModWindow = None
            self.windowStyle = None
            
            self.window = None
            self.ClickOffCloses = True
            self.RefreshWindow()

           
            if isPillars2():
                def getModData(x):
                    return x.ItemModData
                def getModName(x):
                    return getName(getModData(x))
                def getTooltip(x):
                    return trimText(x.GetEffects(StatusEffectFormatMode.Tooltip))
            else:
                def getModData(x):
                    return x.Mod
                def getModName(x):
                    return x.Mod.DisplayName.GetText()
                def getTooltip(x):
                    return trimText(x.Mod.GetTooltipContent(None))
            
            def ShowMinMaxButton(obj, label, attr, min, max, incr=1, refresh=None):
                value = getattr(obj, attr, -1)
                with HorizontalScope():
                    GUILayout.Label(label)
                    GUILayout.FlexibleSpace()
                    GUILayout.Label("(%g)"%(value))
                    if GUILayout.Button("%+g"%(-incr)):
                        newval = value-incr
                        if newval >= min: setattr(obj, attr, newval)
                        if refresh: refresh()
                    if GUILayout.Button("%+g"%(incr)):
                        newval = value+incr
                        if newval <= max: setattr(obj, attr, newval)
                        if refresh: refresh()
                    if GUILayout.Button("Max"):
                        if value <= max: setattr(obj, attr, max)
                        if refresh: refresh()
                        
            def ShowMinMaxButtonFunc(label, getter, setter, min, max, incr=1, refresh=None):
                value = getter()
                with HorizontalScope():
                    GUILayout.Label(label)
                    GUILayout.FlexibleSpace()
                    GUILayout.Label("(%g)"%(value))
                    if GUILayout.Button("%+g"%(-incr)):
                        newval = value-incr
                        if newval >= min: setter(newval)
                        if refresh: refresh()
                    if GUILayout.Button("%+g"%(incr)):
                        newval = value+incr
                        if newval <= max: setter(newval)
                        if refresh: refresh()
                    if GUILayout.Button("Max"):
                        if value <= max: setter(max)
                        if refresh: refresh()
                        
                        
            def ShowMinMaxButtonArray(arr, idx, label, min, max, incr=1, refresh=None):
                value = arr[idx]
                with HorizontalScope():
                    GUILayout.Label(label)
                    GUILayout.FlexibleSpace()
                    GUILayout.Label("(%d)"%(value))
                    if GUILayout.Button("%+d"%(-incr)):
                        newval = value-incr
                        if newval >= min: arr[idx] = newval
                        if refresh: refresh()
                    if GUILayout.Button("%+d"%(incr)):
                        newval = value+incr
                        if newval <= max: arr[idx] = newval
                        if refresh: refresh()
                    if GUILayout.Button("Max"):
                        if value <= max: arr[idx] = max
                        if refresh: refresh()

            def ShowActionAdjustButton(label, callback, value, min, max, incr=1, refresh=None):
                with HorizontalScope():
                    GUILayout.Label(label)
                    GUILayout.FlexibleSpace()
                    GUILayout.Label("(%d)"%(value))
                    if GUILayout.Button("%+d"%(-incr)):
                        callback(-incr)
                        if refresh: refresh()
                    if GUILayout.Button("%+d"%(incr)):
                        callback(incr)
                        if refresh: refresh()
                        
            #UICharacterCreationManager
            def ShowCharacterCreation(refresh=None):
                w = self.window
                c = self.window.ActiveCharacter
                GUILayout.Label(self.windowName)
                stage = int(self.window.CurrentStage)
                stype = int(self.window.CreationType)
                if isPillars2():
                    from Game.GameData import ProgressionTableManagerGameData
                    attrData = ProgressionTableManagerGameData.Instance
                    minv, maxv = attrData.CharacterCreationAttributeMinimum, attrData.CharacterCreationAttributeMaximum
                    attrData = attrData.ProgressionTableManagerComponent
                    totalPoints = "CharacterCreationAttributePoints"
                    skills = [(x, x.GetDisplayName()) for x in c.SkillValues.Keys]
                    SkillValues = c.SkillValues.Item
                else:
                    attrData = self.window
                    totalPoints = "TotalPointBuy"
                    minv, maxv = self.window.StatHardMinimum, self.window.StatHardMaximum
                    skills = [(i,x) for x in enumerate(['Stealth', 'Athletics', 'Lore', 'Mechanics', 'Survival', 'Survival'])]
                    SkillValues = c.SkillValues
                
                if  (stype == 2 and stage == 1) or (stype != 3 and stage == 3): # stats
                    self.state.showStats = GUILayout.Toggle(self.state.showStats, "Base Stats")
                    if self.state.showStats:
                        ShowMinMaxButton(attrData, "Total Points", totalPoints, 0, 52, refresh=refresh)
                        ShowMinMaxButtonArray(c.BaseStats, 1, "Might", minv, maxv, refresh=refresh)
                        ShowMinMaxButtonArray(c.BaseStats, 4, "Constitution", minv, maxv, refresh=refresh)
                        ShowMinMaxButtonArray(c.BaseStats, 2, "Dexerity", minv, maxv, refresh=refresh)
                        ShowMinMaxButtonArray(c.BaseStats, 5, "Perception", minv, maxv, refresh=refresh)
                        ShowMinMaxButtonArray(c.BaseStats, 3, "Intellect", minv, maxv, refresh=refresh)
                        ShowMinMaxButtonArray(c.BaseStats, 0, "Resolve", minv, maxv, refresh=refresh)
                    #self.state.showSkills = GUILayout.Toggle(self.state.showSkills, "Base Skills")
                    #if self.state.showSkills:
                    #    for i, x in skills:
                    #        ShowMinMaxButtonArray(SkillValues, i, x, 0, 20, refresh=refresh)
                    #    ShowMinMaxButton(c, "Skill Points", "SkillPointsToSpend", 0, 20, refresh=refresh)
                #if (stype == 2 and stage == 2): # powers
                #    if GUILayout.Button("AutoGrant Selected"):
                #        w.AutoGrantedAbilities.AddRange(w.GetSelectedAbilities())
                pass
                
            #UICharacterSheetManager
            def ShowCharacterStats(c, refresh=None): #CharacterStats
                GUILayout.Label( getName(c) if isPillars2() else c.Name())
                if isPillars2():
                    csd = c.CharacterStatsData
                    from Game.GameData import ProgressionTableManagerGameData
                    attrData = ProgressionTableManagerGameData.Instance
                    minv, maxv = attrData.CharacterCreationAttributeMinimum, attrData.CharacterCreationAttributeMaximum*2
                else:
                    minv, maxv = 0, 20
                    csd = c
                
                self.state.showStats = GUILayout.Toggle(self.state.showStats, "Base Stats")
                if self.state.showStats:
                    prefix = "_CharacterStatsMutable__m_"                    
                    ShowMinMaxButton(c, "Might", prefix + "baseMight", minv, maxv, refresh=refresh)
                    ShowMinMaxButton(c, "Constitution", prefix + "baseConstitution", minv, maxv, refresh=refresh)
                    ShowMinMaxButton(c, "Dexerity", prefix + "baseDexterity", minv, maxv, refresh=refresh)
                    ShowMinMaxButton(c, "Perception", prefix + "basePerception", minv, maxv, refresh=refresh)
                    ShowMinMaxButton(c, "Intellect", prefix + "baseIntellect", minv, maxv, refresh=refresh)
                    ShowMinMaxButton(c, "Resolve", prefix + "baseResolve", minv, maxv, refresh=refresh)
                    
                self.state.showSkills = GUILayout.Toggle(self.state.showSkills, "Base Skills")
                if self.state.showSkills:
                    if isPillars2():
                        l = c._CharacterStatsMutable__m_skillsList
                        for skill in l.Keys:
                            ShowMinMaxButtonArray(l.Item, skill, getName(skill), 0, 100, refresh=refresh)
                    else:
                        for k in ('Stealth', 'Athletics', 'Lore', 'Mechanics', 'Survival', 'Survival'):
                            ShowMinMaxButton(c, k, k+'Skill', 0, 100, refresh=refresh)
                        ShowMinMaxButton(c, "Skill Points", "RemainingSkillPoints", 0, 20, refresh=refresh)
                    
                if not isPillars2():
                    self.state.showMultipliers = GUILayout.Toggle(self.state.showMultipliers, "Multipliers")
                    if self.state.showMultipliers:
                        ShowMinMaxButton(c, "Attack Speed", "AttackSpeedMultiplier", 0.0, 5.0, 0.25, refresh=refresh)
                        ShowMinMaxButton(c, "Heal Mult", "BonusHealMult", 0.0, 5.0, 0.25, refresh=refresh)
                        ShowMinMaxButton(c, "Hit Crit %", "BonusHitToCritPercent", 0.0, 100.0, 1.0, refresh=refresh)
                        ShowMinMaxButton(c, "Quick Slots", "BonusQuickSlots", 0, 10, 1, refresh=refresh)
                        ShowMinMaxButton(c, "Shield Deflect", "BonusShieldDeflection", 0, 100, 1, refresh=refresh)
                        ShowMinMaxButton(c, "Hit Damage Mult", "CritHitDamageMultiplierBonus", 0.0, 5.0, 0.25, refresh=refresh)
                        ShowMinMaxButton(c, "Damage Min", "DamageMinBonus", 0.0, 20.0, 1.0, refresh=refresh)
                        ShowMinMaxButton(c, "Focus Gain Mult", "FocusGainMult", 0.0, 5.0, 1.0, refresh=refresh)
                        ShowMinMaxButton(c, "Reload Speed Mult", "ReloadSpeedMultiplier", 0.0, 5.0, 1.0, refresh=refresh)
                        ShowMinMaxButton(c, "Rate Of Fire Mult", "RateOfFireMultiplier", 0.0, 5.0, 1.0, refresh=refresh)
                    
                if not isPillars2():
                    self.state.showActions = GUILayout.Toggle(self.state.showActions, "Actions")
                    if self.state.showActions:
                        trait = c.FindFocusTrait()
                        if trait != None: 
                            ShowMinMaxButton(trait, "Focus", "Focus", trait.StartingFocus, trait.MaxFocus)
                        ShowMinMaxButton(c, "Max Health", "MaxHealth", 0, 100)
                        ShowMinMaxButton(c, "Max Stamina", "MaxStamina", 0, 100)
                        mover = c._CharacterStats__m_mover
                        ShowMinMaxButton(mover, "Walk Speed", "WalkSpeed", 0, 10)
                        ShowMinMaxButton(mover, "Run Speed", "RunSpeed", 0, 10)
                        
                        with HorizontalScope():
                            GUILayout.Label('Clear Effects')
                            GUILayout.FlexibleSpace()
                            if GUILayout.Button("All"):
                                c.ClearAllStatusEffects()
                            if GUILayout.Button("Afflictions"):
                                c.ClearEffectsFromAfflictions()
                        with HorizontalScope():
                            GUILayout.Label('Abilities')
                            GUILayout.FlexibleSpace()
                            if GUILayout.Button("Cooldown"):
                                c.RecoveryTimer = 0.0
                                for a in c.ActiveAbilities:
                                    a.RestoreCooldown()
                            if int(c.CurrentFatigueLevel) > 0:
                                if GUILayout.Button("Fatigue"):
                                    c.CurrentFatigueLevel = 0
                        #ShowActionAdjustButton("Adjust Fatigue", c.CurrentFatigueLevel, c.AdjustFatigueLevel, 0, 3)
                    
                pass
                
            def ShowReputation():
                mgr = ReputationManager.Instance
                oldskin = GUI.skin
                try:
                    if self.skin != None: GUI.skin = self.skin

                    self.state.showDisposition = GUILayout.Toggle(self.state.showDisposition, "Disposition")
                    if self.state.showDisposition:
                        if isPillars2():
                            d = mgr.PlayerDisposition.Dispositions
                            for key in d.Keys:
                                ShowMinMaxButtonArray(d.Item, key, getName(key), 0, 100)
                        else:
                            d = mgr.PlayerDisposition.m_dispositions
                            for i, k in enumerate(['Benevolent','Cruel','Clever','Stoic','Aggressive','Diplomatic','Passionate','Rational','Honest','Deceptive']):
                                ShowMinMaxButtonArray(d, i, k, 0, 100)
                        
                    self.state.showFactions = GUILayout.Toggle(self.state.showFactions, "Factions")
                    if self.state.showFactions:
                        options = System.Array[GUILayoutOption]([GUILayout.Width(windowWidth-20), GUILayout.Height(460)])
                        self.state.factionScroll = GUILayout.BeginScrollView(self.state.factionScroll, GUI.skin.scrollView, options)
                        if isPillars2():
                            d = mgr.Reputations
                            for key in d.Keys:
                                rep = d.Item[key]
                                with VerticalScope(getName(key), GUI.skin.window):
                                    ShowMinMaxButton(rep, "Positive", "Positive", 0, 100)
                                    ShowMinMaxButton(rep, "Negative", "Negative", 0, 100)
                        else:
                            for rep in mgr.Factions:
                                with VerticalScope(rep.Name.GetText(), GUI.skin.window):
                                    ShowMinMaxButton(rep, "Positive", "Positive", 0, 100)
                                    ShowMinMaxButton(rep, "Negative", "Negative", 0, 100)
                        GUILayout.EndScrollView()
                        
                    if isPillars2():
                        mgr = PartyManager.Instance
                        rel = mgr.m_companionRelationships
                        
                        options = System.Array[GUILayoutOption]([GUILayout.Width(windowWidth-20), GUILayout.Height(460)])
                        cs = UIInventoryManager.Instance.SelectedObject
                        #pcs = Game.Scripts.TryGetComponentByGuid[CharacterStatsMutable](SpecialCharacterInstanceID.PlayerGuid)
                        #isPlayer = (cs == pcs)
                        cii = ComponentUtils.GetComponent[CompanionInstanceID](cs)
                        isPlayer = (cii == None)

                        self.state.showCompanionDisp = GUILayout.Toggle(self.state.showCompanionDisp, "Companion Disposition")
                        if self.state.showCompanionDisp:
                            self.state.compDispScroll = GUILayout.BeginScrollView(self.state.compDispScroll, GUI.skin.scrollView, options)
                            try:
                        
                                if isPlayer:
                                    for cgd in rel.Keys: # CompanionGameData
                                        pobj = InstanceID.GetObjectByID(cgd.CompanionGuid)
                                        if not mgr.m_activePartyMembersObjects.Contains(pobj): continue
                                        pcs = pobj.GetComponent[CharacterStatsMutable]()
                                        att = rel.Item[cgd] # CompanionAttitude
                                        with HorizontalScope():
                                            GUILayout.Label(getName(pcs))
                                            GUILayout.FlexibleSpace()
                                            GUILayout.Label("(%g)"%(att.GetPlayerRelationshipValue()))
                                            if GUILayout.Button("-1"):
                                                att.AddPlayerRelationship(1, Axis.Negative)
                                            if GUILayout.Button("+1"):
                                                att.AddPlayerRelationship(1, Axis.Positive)
                                                
                                else:
                                    ciid = cii.GetCompanionGuid()
                                    ccgd = [x for x in rel.Keys if ciid == x.CompanionGuid][0]                                    
                                    att = rel.Item[ccgd] # CompanionAttitude
                                    
                                    for cgd in rel.Keys: # CompanionGameData
                                        if ciid == cgd.CompanionGuid: continue
                                        pobj = InstanceID.GetObjectByID(cgd.CompanionGuid)
                                        if not mgr.m_activePartyMembersObjects.Contains(pobj): continue
                                        pcs = pobj.GetComponent[CharacterStatsMutable]()
                                        
                                        with HorizontalScope():
                                            GUILayout.Label(getName(pcs))
                                            GUILayout.FlexibleSpace()
                                            GUILayout.Label("(%g)"%(att.GetRelationshipValue(cgd)))
                                            if GUILayout.Button("-1"):
                                                att.AddRelationship(cgd, 1, Axis.Negative)
                                            if GUILayout.Button("+1"):
                                                att.AddRelationship(cgd, 1, Axis.Positive)
                                                
                            except Exception, ex:
                                GUILayout.Label("Error: " + str(ex))                                    
                            GUILayout.EndScrollView()
                    
                    
                    #isPlayer
                        

                except Exception,ex:
                    GUILayout.Label('Error:' + str(ex))
                GUI.skin = oldskin
                

            def ShowMoney(refresh=None):
                if isPillars2():
                    pinv = GameState.PlayerCharacter.Inventory
                    def getter(): return int(pinv.Money)
                    def setter(x): pinv.Money = int(x)
                    ShowMinMaxButtonFunc("Money", getter, setter, 0, 100000, 1000, refresh=refresh)
                else:
                    pinv = GameState.s_playerCharacter.GetComponent[PlayerInventory]()
                    if pinv: ShowMinMaxButton(pinv.currencyTotalValue, "Money", "v", 0, 100000, 1000, refresh=refresh)

                    
            def ShowCraftingManager():
                mgr = UICraftingManager.Instance
                Recipe.FreeRecipes = GUILayout.Toggle(Recipe.FreeRecipes, "Free Recipes")
                ShowMoney()
                
            def ShowUIAreaMapManager():
                with HorizontalScope():
                    GUILayout.Label('Fog Of War')
                    if GUILayout.Button("Reveal"):
                        fog = self.window._UIAreaMapManager__m_fogOfWar
                        fog.RevealAll()
                        self.window.DoUpdate() if isPillars2() else self.window.Update()

            def ShowUIPartyMember(cs, ai):
                if not cs or not ai:
                    return False
                wasInjured = False
                try:
                    health = ai.GetComponent[Health]()
                    name = getName(cs) if isPillars2() else cs.Name()
                    with VerticalScope(name, GUI.skin.window):
                        with HorizontalScope():
                            if isPillars2():
                                injured = health.CurrentHealthRatio < 1.0
                                wasInjured |= injured
                                if not injured:
                                    if GUILayout.Button("Ok"):
                                        cs.CurrentFatigueLevel = 0
                                        pass
                                elif health.IsUnconscious:
                                    if GUILayout.Button("Wake"):
                                        health.Revive(cs.gameObject)
                                        health.CurrentHealth = health.MaxHealth
                                elif health.IsDead:
                                    if GUILayout.Button("Revive"):
                                        health.Revive(cs.gameObject)
                                        health.CurrentHealth = health.MaxHealth
                                elif health.CountInjuries() > 0:
                                    if GUILayout.Button("Fix"):
                                        for keyword in GlobalReferencesGameData.Instance.InjuryKeywords:
                                            cs.StatusEffectManager.ClearStatusEffectsByKeyword(keyword, True)
                                else:
                                    if GUILayout.Button("Heal"):
                                        health.CurrentHealth = health.MaxHealth
                                GUILayout.FlexibleSpace()
                                
                                trait = cs.GetAccruedResourceTrait(CharacterClass.Cipher)
                                if trait and GUILayout.Button("Focus"):
                                    trait.Focus = trait.GetResourceMax().Value
                                trait = cs.GetAccruedResourceTrait(CharacterClass.Chanter)
                                if trait and GUILayout.Button("Chant"):
                                    trait.PhraseCount = trait.GetResourceMax()
                                trait = cs.GetAccruedResourceTrait(CharacterClass.Monk)
                                if trait and GUILayout.Button('Wounds'):
                                    while not trait.IsResourceAtMax():
                                        trait.AddNewWound()
                                        
                                health._Health__m_canBeTargeted = GUILayout.Toggle(health._Health__m_canBeTargeted, "Target")
                                health.Invulnerable = not GUILayout.Toggle(not health.Invulnerable, "Can Hurt")
                       
                            else:
                            
                                injured = health.HealthPercentage < 0.99 or health.StaminaPercentage < 0.99
                                wasInjured |= injured
                                if not injured:
                                    if GUILayout.Button("Ok"):
                                        cs.CurrentFatigueLevel = 0
                                        pass
                                elif health.Unconscious:
                                    if GUILayout.Button("Wake"):
                                        health.WakeUp()
                                        health.HealthPercentage = 1.0
                                        health.StaminaPercentage = 1.0
                                elif health.Dead:
                                    if GUILayout.Button("Revive"):
                                        health.OnRevive()
                                        health.HealthPercentage = 1.0
                                        health.StaminaPercentage = 1.0
                                else:
                                    if GUILayout.Button("Heal"):
                                        health.HealthPercentage = 1.0
                                        health.StaminaPercentage = 1.0
                                GUILayout.FlexibleSpace()
                                
                                trait = cs.FindFocusTrait()
                                if trait != None: 
                                    if GUILayout.Button("Focus"):
                                        trait.Focus = trait.MaxFocus
                                health.Targetable = GUILayout.Toggle(health.Targetable, "Targetable")
                                health.TakesDamage = GUILayout.Toggle(health.TakesDamage, "Damageable")
                except Exception, e:
                    print(str(e))
                    pass
                return wasInjured
            
            def ShowUIPartyPortrait(refresh=None):
                showAll = False
                for pm, cs, ai in getPartyMembers(minions=True):
                    if cs.RecoveryTimer > 0: showAll = True
                    showAll |= ShowUIPartyMember(cs, ai)
                    
                if showAll:
                    with HorizontalScope():
                        if GUILayout.Button("Heal All"):
                            for p in portraits:
                                Heal(p.SelectedCharacter, p.PartyMemberAI)
                                m = p.Minion
                                if m: Heal(m.SelectedCharacter, m.SelectedPartyMemberAI)
                                    
                        GUILayout.FlexibleSpace()
                        
                        if GUILayout.Button("Afflictions"):
                            for p in portraits:
                                cs = p.SelectedCharacter
                                if cs == None: continue
                                cs.ClearEffectsFromAfflictions()

                                m = p.Minion
                                if m != None:
                                    cs = m.SelectedCharacter
                                    cs.ClearEffectsFromAfflictions()
                            
                        if GUILayout.Button("Cooldown"):
                            for p in portraits:
                                cs = p.SelectedCharacter
                                if cs == None: continue
                                cs.RecoveryTimer = 0.0
                                for a in cs.ActiveAbilities:
                                    a.RestoreCooldown()
                                    
                                m = p.Minion
                                if m != None:
                                    cs = m.SelectedCharacter
                                    cs.RecoveryTimer = 0.0
                                    for a in cs.ActiveAbilities:
                                        a.RestoreCooldown()
                pass

            def InventoryScroll(items, name, posattr, height=300, refresh=None):
            
                oldskin = GUI.skin
                try:
                    if self.skin != None: GUI.skin = self.skin
                    tmpheight = min(height, max(40, items.Count*22+40))
                    
                    #GUILayout.Label('Personal Inventory')
                    options = System.Array[GUILayoutOption]([GUILayout.Width(windowWidth-20), GUILayout.Height(tmpheight)])
                    pos = getattr(self.state, posattr)
                    pos = GUILayout.BeginScrollView(pos, GUI.skin.window, options)
                    try:
                        setattr(self.state, posattr, pos)
                        for item in items:
                            maxcnt = max(2, item.MaxStackSize)
                            def getter(): return int(item.StackSize)
                            def setter(x): item.StackSize = int(x)
                            ShowMinMaxButtonFunc(getName(item), getter, setter, 0, maxcnt, 1, refresh=refresh)
                    finally:
                        GUILayout.EndScrollView()                    
                except Exception,ex:
                    GUILayout.Label('Error:' + str(ex))
                GUI.skin = oldskin
            
            def ShowUIInventoryManager(refresh=None):
                mgr = UIInventoryManager.Instance
                refresh = mgr.DoUpdate if isPillars2() else mgr.Update
                if isPillars2():
                    ShowMoney(refresh=refresh)
                    GUILayout.Space(10)
                    
                    cs = mgr.SelectedObject
                    ai = cs._CharacterStatsMutable__m_aiController
                    p = ai.PartyMember
                    inv = p.Inventory

                    self.state.showPlayerInventory = GUILayout.Toggle(self.state.showPlayerInventory, 'Player Inventory')
                    if self.state.showPlayerInventory: 
                        InventoryScroll(inv, None, 'pinvscrollpos', height=300, refresh=refresh)
                    
                    self.state.showStashInventory = GUILayout.Toggle(self.state.showStashInventory, 'Stash Inventory')
                    if self.state.showStashInventory: 
                        InventoryScroll(inv.StashInventory, None, 'sinvscrollpos', height=300, refresh=refresh)
                    
                else:
                    pinv = GameState.s_playerCharacter.GetComponent[PlayerInventory]()
                    if pinv:
                        ShowMinMaxButton(pinv.currencyTotalValue, "Money", "v", 0, 100000, 1000, refresh=refresh)
                        ShowMinMaxButton(pinv, "Camping Supplies", "campingSupplies", 0, 5, 1, refresh=refresh)
                    GUILayout.Space(10)
                    row = mgr.RowPanel.RowForSelectedPartyMember()
                    for item in row.ItemGrid.ApplicableInventory:
                        if item.GetType() == Equippable: continue
                        ShowMinMaxButton(item, item.BaseItem.DisplayName.GetText(), "StackSize", 0, 10, 1, refresh=refresh)
                pass
                
                
            def ShowItemInspectorWindows():
                itemListWindows = getWindowsOfType(UIItemInspectManager)
                if not itemListWindows: return
                
                for win in itemListWindows:
                    item = win.InspectionObject
                    itemtype = item.GetType()
                    if itemtype == Equippable or itemtype.IsSubclassOf(Equippable):
                        name = getDisplayName(item)
                        with VerticalScope(name, GUI.skin.window):
                            mods = [x for x in item.AttachedItemMods]
                            with HorizontalScope():
                                GUILayout.Label('Mods')
                                GUILayout.FlexibleSpace()
                                if GUILayout.Button('Refresh'):
                                    win.Reload()
                                GUILayout.FlexibleSpace()
                                if GUILayout.Button('Add'):
                                    self.addModWindow = addModToItem(item)
                            for otherWin in itemListWindows:
                                if win == otherWin: continue
                                otherItem = otherWin.InspectionObject
                                if otherItem.GetType() != itemtype: continue
                                otherWin.InspectionObject
                                with HorizontalScope():
                                    GUILayout.Label('Add from ' + getDisplayName(otherItem))
                                    GUILayout.FlexibleSpace()
                                    if GUILayout.Button('Copy'):
                                        for mod in otherItem.AttachedItemMods:
                                            if mod not in mods:
                                                item.AttachItemMod(getModData(mod))
                            
                            for mod in mods:
                                with HorizontalScope():
                                    cc = GUIContent(getModName(mod), getTooltip(mod))
                                    GUILayout.Label(cc)
                                    GUILayout.FlexibleSpace()
                                    GUILayout.Label(str(getModData(mod).Cost))
                                    if GUILayout.Button(GUIContent('-', getTooltip(mod))):
                                        item.DestroyFirstMod(getModData(mod))
                
                options = System.Array[GUILayoutOption]([GUILayout.ExpandHeight(True)])
                GUILayout.Label(GUI.tooltip, options)
                                        
            def ShowUIStrongholdManager():
                ShowMoney()
                    
                with HorizontalScope():
                    GUILayout.Label("Upgrades")
                    GUILayout.FlexibleSpace()
                    if GUILayout.Button("Complete All"):
                        for x in GameState.Stronghold._Stronghold__m_events:
                            if x.EventType == StrongholdEvent.Type.BuildUpgrade:
                                x.Time = 0
                for x in GameState.Stronghold._Stronghold__m_events:
                    if x.EventType == StrongholdEvent.Type.BuildUpgrade:
                        with HorizontalScope():
                            GUILayout.Label(str(x._StrongholdEvent__m_eventUpgradeType))
                            GUILayout.FlexibleSpace()
                            GUILayout.Label("%.1f"%(x.Time / 3600 / 24))
                            if GUILayout.Button("Complete"):
                                x.Time = 0
                                
            # UIShipManagementManager
            def ShowShipManagement():
                ShowMoney()
                cfg = self.window.SelectedObject
                equip = self.window.m_equipment
                
                with HorizontalScope():
                    GUILayout.Label("Add XP")
                    GUILayout.FlexibleSpace()
                    if GUILayout.Button("+100"):
                        cfg.AddExperienceToCrew(100)
                        
                with HorizontalScope():
                    GUILayout.Label("Repair")
                    GUILayout.FlexibleSpace()
                    if cfg.CurrentHullHealth < cfg.GetMaxHullHealth():
                        if GUILayout.Button("Hull"):
                            cfg.CurrentHullHealth = cfg.GetMaxHullHealth()
                    if cfg.CurrentSailHealth < cfg.GetMaxSailHealth():
                        if GUILayout.Button("Sail"):
                            cfg.CurrentSailHealth = cfg.GetMaxSailHealth()
                    if cfg.NeedsRepair:
                        if GUILayout.Button("All"):
                            cfg.CurrentSailHealth = cfg.GetMaxSailHealth()
                            cfg.CurrentHullHealth = cfg.GetMaxHullHealth()
                            
                mgr = ShipCrewManager.Instance
                refresh = mgr.DoUpdate if isPillars2() else mgr.Update
                
                ShowMinMaxButton(mgr, "Morale", "Morale", 0, 100, 10, self.window.DoUpdate)
                with HorizontalScope():
                    GUILayout.Label("Satiate Crew")
                    GUILayout.FlexibleSpace()
                    if GUILayout.Button("Fix"):
                        for m in cfg.GetCrewLayout(None).GetAllCrewMembers():
                            m.DaysWithoutDrink = 0
                            m.DaysWithoutFood = 0
                            m.HasHadDrinkToday = True
                            m.HasHadFoodToday = True
                with HorizontalScope():
                    GUILayout.Label("Heal Crew")
                    GUILayout.FlexibleSpace()
                    if GUILayout.Button("Fix"):
                        inj = SevereInjuryManager.Instance
                        for m in cfg.GetCrewLayout(None).GetAllCrewMembers():
                            inj.RemoveAllInjuries(m.CrewObjectGuid)

                cs = Game.Scripts.TryGetComponentByGuid[CharacterStatsMutable](SpecialCharacterInstanceID.PlayerGuid)
                ai = cs._CharacterStatsMutable__m_aiController
                inv = ai.PartyMember.Inventory
               
                self.state.showFoodInventory = GUILayout.Toggle(self.state.showFoodInventory, 'Food Inventory')
                if self.state.showFoodInventory: 
                    InventoryScroll(inv.ShipFoodInventory, None, 'finvscrollpos', height=300, refresh=refresh)

                self.state.showDrinkInventory = GUILayout.Toggle(self.state.showDrinkInventory, 'Drink Inventory')
                if self.state.showDrinkInventory: 
                    InventoryScroll(inv.ShipDrinkInventory, None, 'dinvscrollpos', height=300, refresh=refresh)

                #self.state.showCrewInventory = GUILayout.Toggle(self.state.showCrewInventory, 'Crew Inventory')
                #if self.state.showCrewInventory: 
                #    InventoryScroll(inv.CrewInventory, None, 'cinvscrollpos', height=300)
                    

                
            def ShowUIStoreManager():
                mgr = self.window
                ShowMoney()
                
            def ShowGlobal():
                with HorizontalScope():
                    self.state.showGlobal = GUILayout.Toggle(self.state.showGlobal, "Players")
                    GUILayout.FlexibleSpace()
                    self.state.showGlobalActions = GUILayout.Toggle(self.state.showGlobalActions, "Actions")
                    GUILayout.FlexibleSpace()
                    self.state.autoCmds = GUILayout.Toggle(self.state.autoCmds, "Auto")
                    
                if self.state.showGlobalActions:
                    with HorizontalScope():
                        GUILayout.FlexibleSpace()
                        if GUILayout.Button("Teleport"):
                            TeleportPartyToDestination()
                        GUILayout.FlexibleSpace()
                        if GUILayout.Button("Heal"):
                            AutoHeal()
                        
                if self.state.showGlobal:
                    ShowUIPartyPortrait()
                    
                if self.state.autoCmds:
                    with HorizontalScope():
                        GUILayout.Label('Enable Auto')
                        self.state.runAutoCmds = GUILayout.Toggle(self.state.runAutoCmds, '')
                        GUILayout.FlexibleSpace()
                        self.state.autoHeal = GUILayout.Toggle(self.state.autoHeal, "Heal")
                    
                if self.state.runAutoCmds and self.counter == 0 and GameState.InCombat:
                    if self.state.autoHeal: AutoHeal()
                pass

            def FuncWindowGUI(windowid):
                self.StartWindowTimer()
                try:
                    if self.skin == None:
                        self.skin = CloneSkin(GUI.skin)
                        self.skin.font = Font( self.skin.font.fontNames, self.skin.font.fontSize-4)
                    
                    pos = Input.mousePosition
                    pos = Vector2(pos.x, UnityEngine.Screen.height-pos.y)
                    
                    if self.window:
                        if self.state.windowRect.Contains(pos):
                            self.window.ClickOffCloses = False
                        elif self.addModWindow != None and self.addModWindow.windowRect.Contains(pos):
                            self.window.ClickOffCloses = False
                        else:
                            self.window.ClickOffCloses = self.ClickOffCloses
                
                    if self.state.minimize:
                        with HorizontalScope():
                            GUILayout.Label(self.windowName)
                            if GUILayout.Button("Restore"):
                                self.state.minimize = False
                                self.state.windowRect = Rect(self.state.windowRect.left-100, self.state.windowRect.top, 
                                    self.state.restoreWindowRect.width, self.state.restoreWindowRect.height)
                                self.state.restoreWindowRect = None

                    else:
                        with HorizontalScope():
                            GUILayout.Label(self.windowName)
                            if GUILayout.Button("Hide (Ctrl+F8)"):
                                self.state.visible = False
                            if GUILayout.Button("Min"):
                                self.state.minimize = True
                                self.state.restoreWindowRect = self.state.windowRect
                                self.state.windowRect = Rect(self.state.windowRect.left+100, self.state.windowRect.top, 150, 50)
                                self.StopWindowTimer()
                            if GUILayout.Button("X"):
                                if self.gameObject:
                                    self.state.visible = False
                                    UnityEngine.Object.DestroyObject(self.gameObject)
                                    return
                                    
                        #if GUILayout.Button("Refresh"):
                        #    self.RefreshWindow()

                        if self.window != None:
                            w = self.window
                            windowType = self.window.Window_Type
                            if isPillars2():
                                CharacterCreation = WindowType.CharacterCreation
                                CharacterSheet = WindowType.CharacterSheet
                                AreaMap = WindowType.AreaMap
                                CharacterCustomizer = WindowType.CharacterCustomizer
                                
                            else:
                                CharacterCreation = WindowType.Character_Creation
                                CharacterSheet = WindowType.Character_Sheet
                                AreaMap = WindowType.Area_Map
                                CharacterCustomizer = WindowType.Character_Customizer
                                
                            self.windowStyle = GUI.skin.window
                            if windowType == CharacterCreation:
                                ShowCharacterCreation(refresh=self.window.RefreshAll)
                            elif windowType == CharacterSheet:
                                cs = w.SelectedObject if isPillars2() else w.SelectedCharacter 
                                #self.window.RefreshCharacter
                                ShowCharacterStats(cs, None)
                                ShowReputation()
                            elif windowType == AreaMap:
                                ShowUIAreaMapManager()
                            elif windowType == CharacterCustomizer:
                                ShowUIInventoryManager()
                                GUILayout.Space(10)
                                ShowItemInspectorWindows()                                
                            elif windowType == WindowType.Inventory:
                                ShowUIInventoryManager()
                                GUILayout.Space(10)
                                ShowItemInspectorWindows()
                            elif windowType == WindowType.Stronghold:
                                ShowShipManagement() if isPillars2() else ShowUIStrongholdManager()
                            elif windowType == WindowType.Store:
                                ShowUIStoreManager()
                            elif windowType == WindowType.Crafting:
                                ShowCraftingManager()
                            else:
                                with HorizontalScope():
                                    GUILayout.Label(str(self.window.Window_Type))
                                    GUILayout.FlexibleSpace()
                                    GUILayout.Label(str(self.window.GetType()))
                                self.windowStyle = GUI.skin.scrollView
                        else:
                            ShowGlobal()
                            self.windowStyle = GUI.skin.scrollView
                            #GUILayout.Label('Nothing Selected')
                            
                    height = Screen.height-self.state.windowRect.top
                    if (self.windowStyle == GUI.skin.window): height -= 10
                    self.state.windowRect = Rect(self.state.windowRect.left, self.state.windowRect.top, self.state.windowRect.width, height)

                    #if Event.current.isMouse:
                    #    if not self.drawWindowRect.Contains(pos):
                    #        Event.current.Use()
                    
                except Exception, e:
                    GUILayout.Label("Error: " + str(e))
                    print(str(e))
                GUI.DragWindow(self.drawWindowRect)
                
                #if Event.current.type == EventType.MouseDown or Event.current.type == EventType.MouseUp:
                    #GUI.FocusControl("")
                    #GUI.FocusWindow(windowid)
                #Event.current.Use()
                
                
            self.windowCallback = GUI.WindowFunction(FuncWindowGUI)
            
        def RefreshWindow(self):
            if self.window != None:
                if not self.window.WindowActive():
                    self.window = None                    
            if self.window == None:
                self.window = getFrontWindow()
                if self.window != None:
                    self.ClickOffCloses = self.window.ClickOffCloses                
            self.windowName = 'Cheats'
            
        def Start(self):
            #self.useGUILayout = True
            if isPillars2():
                import LayerUtility
                self.gameObject.layer = LayerUtility.FindLayerValue(LayerUtility.InGameUILayerName) # set to UI layer
            pass
            
        def OnGUI(self):
            if not self.state.visible: return
            if not self.windowStyle: self.windowStyle = GUI.skin.scrollView
            self.state.windowRect = GUI.Window(1, self.state.windowRect, self.windowCallback, '', self.windowStyle)
            if self.addModWindow != None:
                if self.addModWindow.close:
                    self.addModWindow = None
                else:
                    self.addModWindow.OnGUI()
            
        def Update(self):
            # Update is called less so better place to check keystate
            if Input.GetKeyDown(KeyCode.F8):
                # unity sucks for checking meta keys
                ctrl, alt, shift = unity_util.metakey_state()
                if ctrl and not alt and not shift:
                    self.state.visible = not self.state.visible
                    if self.state.visible: 
                        self.RefreshWindow()
                    else: 
                        self.window = None
                
        # timers dont seem to work in this version of Unity so this doesn't function
        def StartWindowTimer(self):
            self.counter += 1
            if self.counter > 60:
                self.counter = 0
                self.RefreshWindow()
        
            #if not self.timerRunning and self.component:
            #    self.component.StartCoroutine( self.WindowRefreshTimer() )
        
        def StopWindowTimer(self):
            self.timerRunning = False
            
        def WindowRefreshTimer(self):
            self.timerRunning = True
            try:
                self.timerError = None
                while self.state.visible and self.timerRunning:
                    self.counter += 1
                    window = getFrontWindow()
                    if window == None or window != self.window:
                        self.RefreshWindow()
                    yield WaitForSeconds(1)
            except Exception, e:
                self.timerError = "Timer Error: " + str(e)
            self.timerRunning = False
            yield None
    
    return unity_util.create_gui_behavior(Controller)
    
def onSceneChange():
    showWindow(visible=True)        

    